library(tidyverse)
library(ggplot2)
library(ggh4x)
library(ggarchery)
library(stringr)
library(tikzDevice)
INCH.PER.CM <- 1/2.54

WIDTH <- 18.1*INCH.PER.CM
COL.WIDTH <- 8.85*INCH.PER.CM 

BASE.SIZE <- 9
FORMAT <- "tex"

COLOURS.LIST <- c("black", "#E69F00", "#999999", "#009371", "#beaed4", "#ed665a")

options(tikzDocumentDeclaration = "\\documentclass[conference]{IEEEtran}",
        tikzLatexPackages = c(getOption("tikzLatexPackages"),
                              "\\usepackage{amsmath}")
        )

OUTDIR <- "img-tikz/"
do.save.tikz <- function(g, out.name, .width, .height) {
    tikz(str_c(OUTDIR, "/", out.name, ".tex"), width=.width, height=.height)
    print(g)
    dev.off()
}

theme_paper_base <- function() {
    return(theme_bw(base_size=BASE.SIZE) +
           theme(axis.title.x = element_text(size = BASE.SIZE),
                 axis.title.y = element_text(size = BASE.SIZE),
                 legend.title = element_text(size = BASE.SIZE),
                 legend.position = "top", legend.box.spacing = unit(0, "pt"),
                 legend.background = element_rect(fill = "transparent")))
}

############################################################################################
read.bi.data <- function(method) {
    dat <- read_csv(str_c("img/pgfdata/", method, "/guided_spsa.csv"), show_col_types=FALSE) |>
       left_join(read_csv(str_c("img/pgfdata/", method, "/paramter-shift.csv"), show_col_types=FALSE),
                 by=join_by(index)) |>
       rename("Guided SPSA"=`guided-spsa`) |> rename("Parameter Shift"=`parameter-shift`) |>
       pivot_longer(cols=!(index))
    
    dat$lw <- "0"
    dat[dat$name=="Guided SPSA",]$lw="1"
    dat$lw <- as.factor(dat$lw)

    return(dat)
}

dat <- read.bi.data("classification")
    
g <- ggplot(dat, aes(x=index, y=value, colour=name)) + geom_line(aes(linewidth=lw)) +
    theme_paper_base() + xlab("Epoch") + ylab("Average Mean Square Error") +
    scale_colour_manual("", values=COLOURS.LIST) +
    scale_discrete_manual("linewidth", values=c(0.5, 1.0)) + guides(linewidth="none")

do.save.tikz(g, "convergence_grad_ideal_iris", COL.WIDTH, 0.65*COL.WIDTH)

#################
dat <- read.bi.data("regression/subopt")

g <- ggplot(dat, aes(x=index, y=value, colour=name)) + geom_line(aes(linewidth=lw)) +
    theme_paper_base() + xlab("Epoch") + ylab("Average Mean Square Error") +
    scale_colour_manual("", values=COLOURS.LIST) +
    scale_discrete_manual("linewidth", values=c(0.5, 1.0)) + guides(linewidth="none")

do.save.tikz(g, "convergence_grad_zero_init", COL.WIDTH, 0.65*COL.WIDTH)


#################
combinations  <- as_tibble(expand.grid(c("gspsa", "ps"), c("random", "zero")))

dat <- do.call(rbind, lapply(1:nrow(combinations), function(i) {
    dat <- read_csv(str_c("img/pgfdata/regression/subopt/histogram/",
                          str_c(combinations[i,] |>
                                reframe(combination=str_c(Var1, Var2, sep="_"))), ".csv"),
                    show_col_types=FALSE)
    dat$algo <- combinations[i,]$Var1
    dat$variant <- combinations[i,]$Var2

    return(dat)
})) |> pivot_longer(cols=1:20) |>
       group_by(algo, variant) |> mutate(value.norm=value/max(value), value2=as.integer(value/max(value)*100)) |>
       ungroup() |> mutate(epoch=as.integer(name)) |> mutate(epoch.label="Epoch") |>
       mutate(algvar=str_c(algo, "/", variant)) |> filter(epoch %in% c(0,5,10,15,19))

labels.algo <- c(ps="Parameter Shift", gspsa="Guided SPSA")
labels.variant <- c(random="Random", zero="Zero")

g <- ggplot(dat |> uncount(value2), aes(x=bin_centers)) +
    #geom_bar(stat="identity") +
    geom_density() +
    facet_nested(cols=vars(epoch.label, epoch), rows=vars(algo, variant),
                 scales="free_y", labeller=labeller(algo=labels.algo,
                                                    variant=labels.variant)) +
    theme_paper_base() + xlab("Gradient Value") + ylab("Relative Frequency [a.u.]")

g <- ggplot(dat, aes(x=bin_centers, y=value.norm, colour=algvar)) +
    geom_bar(stat="identity") +
    facet_nested(cols=vars(epoch.label, epoch), rows=vars(algo, variant),
                 scales="free_y", labeller=labeller(algo=labels.algo,
                                                    variant=labels.variant)) +
    theme_paper_base() + scale_x_continuous("Gradient Value", breaks=seq(-0.06, 0.06, 0.03)) +
    guides(color="none") + ylab("Relative Frequency [a.u.]") + scale_colour_manual("", values=COLOURS.LIST) +
    theme(axis.text.x=element_text(angle=45, hjust=1))

do.save.tikz(g, "gradient_distribution", WIDTH, 0.85*COL.WIDTH)

##############################
dat.func <- read_csv("img/pgfdata/toy_problem/function.csv", show_col_types=FALSE)
dat.ps <- read_csv("img/pgfdata/toy_problem/parameter-shift.csv", show_col_types=FALSE)
dat.gspsa <- read_csv("img/pgfdata/toy_problem/guided_spsa.csv", show_col_types=FALSE)

segmentise.data <- function(dat, max.idx) {
    return(dat |> filter(index < max.idx) |> mutate(xend=lead(x), yend=lead(y)) |>
           filter(!is.na(xend)) |>
           mutate(arrow_x = (x + xend) / 2, arrow_y = (y + yend) / 2))
}

dat <- rbind(segmentise.data(dat.gspsa, 25) |> mutate(algo="Guided SPSA"),
             segmentise.data(dat.ps, 20) |> mutate(algo="Parameter Shift"))

##                 arrow=arrow(length=unit(0.5, "mm")), arrow.fill=COLOURS.LIST[[2]], type="closed") +

g <- ggplot(dat, aes(x=x, y=y)) + geom_line(dat=dat.func, colour=COLOURS.LIST[[1]]) +
    geom_arrowsegment(aes(xend=xend, yend=yend), linewidth=0.25,
                      colour=COLOURS.LIST[[2]], fill=COLOURS.LIST[[2]],
                      arrows=arrow(type="closed", length=unit(0.75, "mm"))) +
    theme_paper_base() + facet_grid(cols=vars(algo)) +
    xlab("Model Output") + ylab("Cost") 

do.save.tikz(g, "convergence_path", COL.WIDTH, 0.65*COL.WIDTH)

#########################################
variants <- c("paramter-shift"="Parameter Shift",
              "spsa_10"="SPSA 10", "spsa_20"="SPSA 20",
              "spsa_30"="SPSA 30", "spsa_40"="SPSA Max",
              "guided_spsa"="Guided SPSA")

read.experiment <- function(experiment, .colnames=c("index", "value")) {
    dat <- do.call(rbind, lapply(names(variants), function (variant) { 
        file <- str_c("img/pgfdata/regression/", experiment, "/",
                      variant, ".csv")
        dat <- read_csv(file, show_col_types=FALSE)
        colnames(dat) <- .colnames
        dat$variant <- variant
        
        return(dat)
    }))

    dat$lw <- "0"
    dat[dat$variant=="guided_spsa",]$lw="1"
    dat$lw <- as.factor(dat$lw)

    return(dat)
}

dat <- read.experiment("ideal/friedmann1")

g <- ggplot(dat, aes(x=index, y=value, colour=variant)) +
    geom_line(aes(linewidth=lw)) +
    theme_paper_base() + scale_colour_manual("", values=COLOURS.LIST,
                                             labels=variants) +
    scale_discrete_manual("linewidth", values=c(0.5, 1.0)) +
    guides(color=guide_legend(nrow=2, byrow=TRUE), linewidth="none") +
    theme(legend.key.spacing.y=unit(-2, "mm")) +
    xlab("Epoch") + scale_y_sqrt("Avg.~Mean Square Error [sqrt]", breaks=c(0.06, seq(0.1, 0.3, 0.05)))

do.save.tikz(g, "friedman_gradients_ideal", COL.WIDTH, 0.65*COL.WIDTH)


##############################################
dat <- read.experiment("ideal/boston")

g <- ggplot(dat, aes(x=index, y=value, colour=variant)) +
    geom_line(aes(linewidth=lw)) +
    theme_paper_base() + scale_colour_manual("", values=COLOURS.LIST,
                                             labels=variants) +
    scale_discrete_manual("linewidth", values=c(0.5, 1.0)) +
    guides(color=guide_legend(nrow=2, byrow=TRUE), linewidth="none") +
    theme(legend.key.spacing.y=unit(-2, "mm")) +
    xlab("Epoch") + scale_y_sqrt("Avg.~Mean Square Error [sqrt]", breaks=seq(0.15, 0.45, 0.05))

do.save.tikz(g, "boston_gradients_ideal", COL.WIDTH, 0.65*COL.WIDTH)


#############################################
dat <- read.experiment("CDF", c("index", "value", "bins"))

g <- ggplot(dat, aes(x=bins, y=value, colour=variant)) +
    geom_line(aes(linewidth=lw)) +
    theme_paper_base() + scale_colour_manual("", values=COLOURS.LIST,
                                             labels=variants) +
    scale_discrete_manual("linewidth", values=c(0.5, 1.0)) +
    guides(color=guide_legend(nrow=2, byrow=TRUE), linewidth="none") +
    theme(legend.key.spacing.y=unit(-2, "mm")) +
    scale_x_continuous("Bins (MAE)", breaks=seq(0.05, 0.35, 0.05)) + scale_y_continuous("Count")

do.save.tikz(g, "friedman_sample_cdf", COL.WIDTH, 0.65*COL.WIDTH)


##########################
dat <- read_csv("img/pgfdata/reinforcement_learning/guided_spsa.csv") |>
       mutate(type="Guided SPSA") |>
       rbind(read_csv("img/pgfdata/reinforcement_learning/paramter-shift.csv") |>
             mutate(type="Parameter Shift Rule")) |>
       rename("index"=1)
dat$colour <- factor("Mean", levels=c("Minimum/Maximum", "Mean"))
dat2 <- dat
dat2$colour <- factor("Minimum/Maximum", levels=c("Minimum/Maximum", "Mean"))

g <- ggplot(dat, aes(x=index, y=val_mean, colour=colour)) + theme_paper_base() +
    geom_linerange(data=dat2, aes(x=index, ymin=val_min, ymax=val_max), alpha=0.35) +
    scale_colour_manual("", values=c(COLOURS.LIST[[2]], COLOURS.LIST[[1]])) +
    geom_line() + xlab("Epoch") + ylab("Reward") + facet_grid(cols=vars(type))
print(g)

do.save.tikz(g, "qrl", COL.WIDTH, 0.65*COL.WIDTH)
